function [out] = canny(in,w_size,Thresh)

%   Canny edge detection 
%       EDGE = CANNY(X) returns the edge map of image X with respect to Canny's 
%       method. 
%
%       EDGE = CANNY(X,N) uses a NxN processing window (default: N=3).
%
%       EDGE = CANNY(X,N,THRESH) uses the predefined values in THRESH as thresholds.
%       THRESH is a two-element vector than contains the high and the low thresholds used
%       in the hysterezis thresholding step. If THRESH contains one single value, then this
%       value is set to the high threshold.

if ~exist('w_size','var')
    w_size = 3;
end

if ~exist('in','var')
    help canny
    error('You must specify an input image!')
end

w_size = floor(w_size/2);

[lin col] = size(in);

%Smoothing gaussian filter
disp('Performing Gaussian smoothing.')
gauss = exp(-(-w_size:w_size)'.^2/(2*w_size^2))*exp(-(-w_size:w_size).^2/(2*w_size^2));
in = conv2(in,gauss/sum(gauss(:)),'same');

%Computing the coefficients of the Sobel operator
disp('Computing preliminary gradient using the Sobel operator.');
pasc_tri = pascal(2*w_size+1);
pasc_tri = pasc_tri(:,(2*w_size+1):-1:1);
smooth = diag(pasc_tri);
deriv = diag(pasc_tri,1);
deriv = [deriv;0]-[0;deriv];
sobel_y = conv2(deriv,smooth','full');
sobel_x = -sobel_y';

gy = conv2(in,sobel_y,'same');
gx = conv2(in,sobel_x,'same');

magn = sqrt(gx.^2+gy.^2);

cadran = atan2(gy,gx);
cadran = cadran+pi.*(cadran<0);
cadran = 1+min(4,floor(cadran/(pi/4)));

gx = abs(gx);
gy = abs(gy);

%Non-maximum suppression

disp('Performing non-maximum suppression.')
magn_supp = zeros(lin,col);
for i = 2:lin-1,
    for j = 2:col-1
        if magn(i,j)>0,
            switch cadran(i,j),
            case 1,
                m1 = gy(i,j)/gx(i,j)*magn(i,j+1)+(1-gy(i,j)/gx(i,j))*magn(i-1,j+1);
                m2 = gy(i,j)/gx(i,j)*magn(i,j-1)+(1-gy(i,j)/gx(i,j))*magn(i+1,j-1);
            case 2,
                m1 = gx(i,j)/gy(i,j)*magn(i-1,j)+(1-gx(i,j)/gy(i,j))*magn(i-1,j+1);
                m2 = gx(i,j)/gy(i,j)*magn(i+1,j)+(1-gx(i,j)/gy(i,j))*magn(i+1,j-1);                
            case 3,
                m1 = gx(i,j)/gy(i,j)*magn(i-1,j)+(1-gx(i,j)/gy(i,j))*magn(i-1,j-1);
                m2 = gx(i,j)/gy(i,j)*magn(i+1,j)+(1-gx(i,j)/gy(i,j))*magn(i+1,j+1);                
            case 4,
                m1 = gy(i,j)/gx(i,j)*magn(i,j-1)+(1-gy(i,j)/gx(i,j))*magn(i-1,j-1);
                m2 = gy(i,j)/gx(i,j)*magn(i,j+1)+(1-gy(i,j)/gx(i,j))*magn(i+1,j+1);                
            end
            if all(magn(i,j)>=[m1 m2])
                magn_supp(i,j)=magn(i,j);
            end
        end
    end
end
disp('Now performing hysterezis thresholding.');

if ~exist('Thresh','var'),
    [h,b] = hist(magn_supp(:),100);
    h = cumsum(h/sum(h));
    T_high = b(min(find(ceil(h-.95))));
    T_low = b(min(find(ceil(h-.9))));
elseif length(Thresh) == 1,
    T_high = Thresh;
    T_low = .6*T_high;
end
    
disp(['TH = ' num2str(T_high) '; TL = ' num2str(T_low)]);

out = magn_supp >= T_high;
[ind_lin,ind_col]=find(out);

map = zeros(lin,col);
sir = zeros(10000,2);

for k = 1:size(ind_lin,1),
    if ~map(ind_lin(k),ind_col(k))
        %Region growing
        map(i,j)=1;
        sir(1,1)=ind_lin(k);
        sir(1,2)=ind_col(k);
        ptr=2;
        ptr1=1;
        while (ptr1<ptr)
            i=sir(ptr1,1);
            j=sir(ptr1,2);
            for m=-1:1,
                for n=-1:1,     
                    if i+m >=1 & i+m <= lin & j+n >=1 & j+n <= col,
                        if ~map(i+m,j+n),
                            if magn_supp(i+m,j+n) >= T_low,
                                sir(ptr,:)=[i+m,j+n];
                                map(i+m,j+n)=1;
                                ptr=ptr+1;
                            end
                        end
                    end
                end
            end         
            ptr1=ptr1+1;
        end
        sir=sir(1:(ptr-1),:);
        %Region growing done
        sir = lin*(sir(:,2)-1)+sir(:,1);
        out(sir) = ones(size(sir));
    end
end

disp('DONE!');


